import re
import time
from abc import ABC
from contextlib import suppress
from dataclasses import dataclass
from enum import Enum
from typing import Callable, Any, Tuple, List, Union, Dict

import pytest
from _pytest.mark import Mark
from junit_xml import TestCase


class TestCaseCategories(Enum):
    SUITE = "suite-function"
    FUNCTION = "function"
    FIXTURE = "fixture"
    FIXTURE_TEARDOWN = "fixture-teardown"


@dataclass
class CaseFailure:
    message: str
    output: str = ""
    type: str = ""

    def __getitem__(self, item):
        return self.__getattribute__(item)


@dataclass
class TestCaseData:
    case: TestCase
    _func: Callable
    _start_time: float
    parametrize: Union[None, List[Tuple[str, Any]]] = None
    _has_parent = False

    @property
    def name(self):
        return self._func.__name__

    @property
    def has_parent(self):
        return self._has_parent

    def set_fin_time(self):
        self.case.elapsed_sec = time.time() - self._start_time

    def get_case_key(self):
        """ return immutable key """
        if self.parametrize:
            return tuple(sorted(self.parametrize))
        return ()

    def set_parametrize(self, params: List[Tuple[str, Any]]):
        self.parametrize = params
        self.case.name += f'[{", ".join([f"{p[0]}={p[1]}" for p in params])}]'

    def set_parent(self, case_parent_name: str):
        self._has_parent = True
        self.case.classname = f"{self.case.classname}.{case_parent_name}"


class Utils(ABC):
    JUNIT_EXCEPTION_TAG = "__is_junit_exception__"

    @staticmethod
    def get_new_test_case(func: Callable, classname: str, category: TestCaseCategories) -> TestCase:
        return TestCase(name=func.__name__, classname=classname, category=category.value)

    @classmethod
    def is_case_exception_already_raised(cls, exception: BaseException) -> bool:
        return hasattr(exception, cls.JUNIT_EXCEPTION_TAG)

    @classmethod
    def mark_case_exception_as_raised(cls, exception: BaseException):
        setattr(exception, cls.JUNIT_EXCEPTION_TAG, True)

    @staticmethod
    def get_wrapped_function(func: pytest.Function, func_name: str = None) -> Callable:
        """
        Get wrapped function from class or module
        :param func: wrapped function from class or module
        :param func_name: If there is pytest parameterized suite name is needed
        :return:
        """
        if func_name is None:
            func_name = func.name

        if func.cls:
            return getattr(func.cls, func_name).__wrapped__
        return getattr(func.module, func_name).__wrapped__


class PytestUtils:
    PARAMETERIZED_KEY = "parametrize"
    OWN_MARKERS_KEY = "own_markers"

    @classmethod
    def get_case_pytest_parameterized(cls, pytest_function: pytest.Function) -> List[Tuple[str, Any]]:
        return (list(
            sorted(pytest_function.callspec.params.items())) if pytest_function and pytest_function.own_markers and all(
            m.name == cls.PARAMETERIZED_KEY for m in pytest_function.own_markers) else list())

    @classmethod
    def get_suite_pytest_parameterized(cls,
                                       cases_data: List[TestCaseData],
                                       func: Callable,
                                       stack_locals: List[Dict[str, Any]]
                                       ) -> List[Tuple[str, Any]]:
        parameterized = list()
        if len(cases_data) == 0:
            return cls._get_parameterized_on_no_cases(func, stack_locals)
        test_cases = [test_case_data for test_case_data in cases_data if test_case_data.parametrize]
        if test_cases:
            parameterized = [k for k in set([c.get_case_key() for c in test_cases if len(c.get_case_key()) > 0]).pop()]
        return parameterized

    @classmethod
    def _get_parameterized_on_no_cases(cls, func: Callable, stack_locals) -> List[Tuple[str, Any]]:
        for f_locals in [stack_local for stack_local in stack_locals if "pyfuncitem" in stack_local]:
            pytest_func = f_locals["pyfuncitem"]
            if hasattr(func, "pytestmark"):
                parameterized = [m.args[0] for m in func.pytestmark if m.name == cls.PARAMETERIZED_KEY]
                if isinstance(pytest_func, pytest.Function) and parameterized:
                    return sorted(list({k: v for k, v in pytest_func.funcargs.items() if k in parameterized}.items()))

        return list()

    @classmethod
    def get_fixture_suite(cls, func: pytest.Function):
        with suppress(AttributeError):
            # if pytest.mark.parametrize exist, get actual function from class while ignoring the add parameters
            if hasattr(func, cls.OWN_MARKERS_KEY) and len(func.own_markers) > 0:
                mark_function = cls.get_fixture_mark_function(func)
                if mark_function:
                    return mark_function
            return Utils.get_wrapped_function(func)

        return None

    @classmethod
    def get_fixture_data(cls, func: pytest.Function) -> Tuple[List[str], List[Mark]]:
        marks = [m for m in func.own_markers if m.name == cls.PARAMETERIZED_KEY]
        marks_count = len(marks)

        if marks_count == 0:
            return [], marks

        params_regex = "-".join(["(.*?)"] * marks_count)
        args = list(re.compile(r"(.*?)\[{0}]".format(params_regex)).findall(func.name).pop())
        return args, marks

    @classmethod
    def get_fixture_parameterized(cls, func: pytest.Function) -> List[Tuple]:
        args, marks = cls.get_fixture_data(func)
        marks_count = len(marks)
        if not marks:
            return []
        args.pop(0)
        parameterized = list()
        for i in range(marks_count):
            parameterized.append((marks[i].args[0], args[i]))
        return parameterized

    @classmethod
    def get_fixture_mark_function(cls, func: pytest.Function) -> Union[Callable, None]:
        """
        If mark parameterize decorate test suite with given fixture _get_mark_function is searching
        for all parametrize arguments and permutations.
        If mark type is parametrize, set self._parametrize with the current argument permutation.
        The representation of function name and arguments for N parametrize marks are as followed:
            func.name = some_func_name[param1-param2-...-paramN]
        :param func: wrapped function contained with pytest Function object
        :return: wrapped function itself
        """
        args, marks = cls.get_fixture_data(func)
        marks_count = len(marks)
        if not marks:
            return None

        func_name = args.pop(0)

        parameterized = list()
        for i in range(marks_count):
            parameterized.append((marks[i].args[0], args[i]))

        return Utils.get_wrapped_function(func, func_name)



from types import GeneratorType
from typing import Callable, Union

import decorator

from ._junit_test_case import JunitTestCase, TestCaseCategories
from ._junit_test_suite import JunitTestSuite
from .utils import Utils


class JunitFixtureTestCase(JunitTestCase):
    """
    This class defines TestCase for pytest fixtures.
    When use, need to make sure that the fixture come first, for example:
        @pytest.fixture
        @JunitFixtureTestCase()
        def my_fixture()
            ...
    """

    _generator: Union[GeneratorType, None]
    _inner_test_case_exception: bool

    def __init__(self) -> None:
        super().__init__()
        self._generator = None
        self._inner_test_case_exception = False

    def __call__(self, function: Callable) -> Callable:
        self._func = function

        def wrapper(_, *args, **kwargs):
            value = self._wrapper(function, *args, **kwargs)
            yield value
            try:
                if self._generator:
                    self._teardown_yield_fixture(self._generator)
            except BaseException as e:
                self._case_data.case.category = TestCaseCategories.FIXTURE_TEARDOWN.value
                self._add_failure(e)
                raise
            finally:
                JunitTestSuite.fixture_cleanup(self._case_data, self.get_suite_key())

        return decorator.decorator(wrapper, function)

    @classmethod
    def _teardown_yield_fixture(cls, it) -> None:
        """Execute the teardown of a fixture function by advancing the iterator
        after the yield and ensure the iteration ends (if not it means there is
        more than one yield in the function)."""
        if type(it) != GeneratorType:
            # Nothing needs to be done for non-generator fixtures
            return

        try:
            next(it)
        except StopIteration:
            pass

    def _execute_wrapped_function(self, *args, **kwargs):
        fixture_ret = super()._execute_wrapped_function(*args, **kwargs)

        if type(fixture_ret) == GeneratorType:
            # This is likely a pytest fixture that uses the `yield` teardown pattern
            generator = fixture_ret

            if generator.__name__ == "<genexpr>":
                # This is actually* a regular generator expression the user is trying to return, not a 
                # pytest generator function used for setup / teardown. Just return it without special
                # handling.
                # * (Generator functions have their `__name__` attribute equal to the actual function name)
                return generator

            # TODO: What if the user actually wants to return a generator function from their fixture?
            #       Find a way to tell apart `yield` for teardown vs user actually returning a generator function.

            try:
                self._generator = generator
                return next(generator)
            except (StopIteration, TypeError):
                return None
            except BaseException as e:
                if Utils.is_case_exception_already_raised(e):
                    self._inner_test_case_exception = True
                raise
        else:
            # This is a regular fixture that simply returns a value
            return fixture_ret

    def _on_wrapper_end(self):
        """
        Fixtures executing before the test suite, due to that issue the suite can't collect the fixture case
        if there is an exception during it's execution. If exception occur, it call JunitTestSuite class method
        collect_all that trigger the suite to collect all cases and export them into xml
        :return: None
        """
        self._case_data.case.category = TestCaseCategories.FIXTURE.value

        super(JunitFixtureTestCase, self)._on_wrapper_end()
        if len(self._case_data.case.failures) > 0 or self._inner_test_case_exception:
            JunitTestSuite.collect_all(True)



import traceback
from contextlib import suppress
from typing import Any, Callable, Dict, List, Union

import pytest

from ._junit_decorator import JunitDecorator
from ._junit_test_suite import JunitTestSuite
from .utils import TestCaseCategories, TestCaseData, CaseFailure, Utils, PytestUtils


class JunitTestCase(JunitDecorator):
    """
    JunitTestCase is a decorator that represents single TestCase.
    When the decorated function finished its execution, it's registered to the relevant JunitTestSuite.
    TestCase will fail (TestCase failure) only when exception occurs during execution.
    """

    _stack_locals: List[Dict[str, Any]]
    _case_data: Union[TestCaseData, None]
    _pytest_function: Union[pytest.Function, None]

    def __init__(self) -> None:
        super().__init__()
        self._stack_locals = list()
        self._case_data = None
        self._pytest_function = None

    @property
    def name(self):
        return self._case_data.name if self._case_data.name else self._func.__name__

    def _on_wrapper_start(self, function):
        super()._on_wrapper_start(function)
        case = Utils.get_new_test_case(function, self._get_class_name(), TestCaseCategories.FUNCTION)
        self._case_data = TestCaseData(_start_time=self._start_time, case=case, _func=function)
        self._pytest_function = [
            stack_local for stack_local in self._stack_locals
            if "self" in stack_local and isinstance(stack_local["self"], pytest.Function)][0]["self"]

    def _add_failure(self, e: BaseException, message_prefix: str = ""):
        message = f"{message_prefix} {str(e)}" if message_prefix else str(e)
        failure = CaseFailure(message=message, output=traceback.format_exc(), type=e.__class__.__name__)
        self._case_data.case.failures.append(failure)

    def _on_exception(self, e: BaseException):
        if Utils.is_case_exception_already_raised(e):
            raise  # already registered on son test case
        Utils.mark_case_exception_as_raised(e)
        self._add_failure(e)
        raise e

    def _on_wrapper_end(self):
        self._case_data.set_fin_time()
        JunitTestSuite.register_case(self._case_data, self.get_suite_key())

    def get_suite_key(self) -> Union[Callable, None]:
        """
        Get suite function as unique key.
        This function handles also on case that the test suite function decorated with pytest.mark.parametrize and
        collect its parameters and record them into the test case.
        :return: Wrapped Suite function instance
        """
        with suppress(AttributeError):
            suite_func = self.get_suite()
            return suite_func

        return None

    def _set_parameterized(self, parameterized: List = None):
        if not parameterized:
            parameterized = PytestUtils.get_case_pytest_parameterized(self._pytest_function)

        if parameterized and not self._case_data.parametrize:
            self._case_data.set_parametrize(parameterized)

    def get_suite(self):
        stack_locals = [stack_local for stack_local in self._stack_locals if "self" in stack_local]
        function_locals = [stack_local for stack_local in stack_locals if "function" in stack_local]

        self._set_parameterized()
        is_inside_fixture = False

        suite_func = self._get_function_suite(function_locals)
        if suite_func is None:
            is_inside_fixture = True
            suite_func = PytestUtils.get_fixture_suite(self._pytest_function)
            fixture_params = PytestUtils.get_fixture_parameterized(self._pytest_function)
            if fixture_params:
                self._set_parameterized(fixture_params)

        self._case_data.is_inside_fixture = is_inside_fixture
        return suite_func

    def _get_function_suite(self, function_locals: List[Dict[str, Any]]):
        suite_func = None

        for f_locals in function_locals:
            suite = f_locals["self"]
            if isinstance(suite, JunitTestCase) and f_locals["function"].__name__ != self.name:
                if not self._case_data.has_parent:
                    self._case_data.set_parent(f_locals["function"].__name__)

            if isinstance(suite, JunitTestSuite):
                suite_func = f_locals["function"]
                break
        return suite_func
